//***************************************************************************
//  File........: main.c
//  Author(s)...: Chiper
//  URL(s)......: http://digitalchip.ru/
//  Device(s)...: ATMega8-8AU
//  Frequency...: 2MHz
//  Compiler....: AVR-GCC
//  Description.: Bathroom fan controller
//  Data........: 22.07.2013
//  Version.....: 1.1
//***************************************************************************
//  Notice: Not optimized
//  
//***************************************************************************

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>

///////////////////////////////
// DEFINITIONS
///////////////////////////////

#define SENSOR_DDR  DDRB		// Порт, который отвечает за сенсор
#define SENSOR_PORT PORTB
#define SENSOR_PIN	PINB
#define SENSOR_A 	1			// PIN1 (positive) (анод)
#define SENSOR_K 	2			// PIN2 (negative) (катод)

#define MECH_DDR  	DDRC		// Порт, который отвечает за исполнительный механизм (вентилятор)
#define MECH_PORT 	PORTC
#define MECH_PIN1 	3			// 2 пина для умощнения выхода
#define MECH_PIN2 	2

#define LIGHT_THRESHOLD	200		// Порог срабатывания светодиода как датчика (подбирается индивидуально
								// по условиям освещенности в месте, где установлен датчик)

#define BLINK_SPEED	25			// Скорость мигания светодиода, как светодиода

#define TCNT0_BASE 	0x83		// Основание для расчета времени на Timer0 (1 тик - 500 микросекунд) (управление нагрузкой)
#define TCNT2_BASE 	0x06		// Основание для расчета времени на Timer2 (1 тик - 1 миллисекунда)  (системный таймер)

#define LOAD_ON		1			// Нагрузка включена
#define LOAD_OFF	0			// Нагрузка выключена

// Макросы для работы с битами
#define ClearBit(reg, bit)       reg &= (~(1<<(bit)))
#define SetBit(reg, bit)         reg |= (1<<(bit))
#define InvBit(reg, bit)         reg ^= 1<<bit

// GLOBAL VARIABLES
volatile uint8_t LOAD_STATUS = LOAD_OFF;				// Состояние нагрузки

volatile uint8_t STATE_CTRL = 0;						// Вспомогательная переменная для работы с нагрузкой

volatile uint8_t blink_count = 0;						// Счетчик для задержки мигания светодиода

volatile uint32_t sys_time = 0;							// Системный таймер
volatile uint32_t sensor_timer1, sensor_timer2  = 0;	// Время включения и выключения датчика
volatile uint32_t load_timer  = 0;						// Таймер, в течении которого включена нагрузка
volatile uint8_t  sensor_started_flag = 0;				// Флаг, показывающий, что сенсор уже сработал ранее

// FUNCTION PROTOTYPES
void setupHW(void);										// Инициализация
uint8_t GetSensorState(void);							// Читаем состояние сенсора
void SensorBlink(void);									// Мигаем светодиодом с учетом задержек

///////////////////////////////
// MAIN PROGRAMM
///////////////////////////////
int main(void)
{
	//	***** SETUP
	setupHW();

	//  ***** LOOP  (бесконечная шарманка)
	while (1)
	{
		if ( GetSensorState() ){	// Если сработал датчик
			if (sensor_started_flag == 0){ 	// и ранее он не срабатывал
				sensor_timer1 = sys_time;	// засекаем время срабатывания датчика
				sensor_started_flag = 1;	// устанавливаем флаг, что датчик сработал
			}
			else {							// но ранее он уже сработал
				sensor_timer2 = sys_time;	// засекаем время выключения датчика (шарманка бесконечная, поэтому
											// это время постоянно обновляется, пока не выключится датчик)

				if ((sensor_timer2 - sensor_timer1)>15000) load_timer = 120000;	//задержка в 15 сек, перед включением ветилятора
																				//затем работает 2 минуты, если никто не изменит это время
			}
		}
		else				// Если датчик не работает
		{
			if(sensor_started_flag){	// Если раньше датчик срабатывал

				uint32_t sel_time = sensor_timer2 - sensor_timer1;	// Время, в течении которого датчик показывал уровень выше порога

//!!! не оптимизировано
				if (sel_time>300000)  load_timer = 180000;	// если больше 5' то включить нагрузку на 3'
				if (sel_time>600000)  load_timer = 360000;	// если больше 10' то включить нагрузку на 6'
				if (sel_time>900000)  load_timer = 540000;	// если больше 15' то включить нагрузку на 9'
				if (sel_time>1200000) load_timer = 720000;	// если больше 20' то включить нагрузку на 12'
				if (sel_time>1500000) load_timer = 900000;	// если больше 25' то включить нагрузку на 15'

				sensor_started_flag = 0;	// Флаг, что датчик обработан и сенсор не срабатывает
			}
			else	// датчик выключен и ранее не срабатывал
			{
				if (load_timer == 0){	// Ждем, пока отработает нагрузка положенное время
					sys_time = 0;		// Обнуляем систменые переменные, что бы к началу срабатывания датчика
					sensor_timer1 = 0;	// у нас было свежее состояние таймеров
					sensor_timer2 = 0;
				}

			}
		}

		SensorBlink();	// Мигаем светодиодом-сенсором

		_delay_ms(100);	// Задержка бесконечной шарманки, что б не сильно часто молотила и было видно
						// включение светодиода-сенсора как светодиода

	}	// (конец шарманки)
}

////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////

///////////////////////////////
// FUNCTIONS
///////////////////////////////

// Инициализация периферии и переменных
void setupHW(void)
{
	// *** Запрещаем прерывания на время настройки
	cli();

	// *** Настройка GPIO
	SENSOR_DDR |= (1<<SENSOR_A) | (1<<SENSOR_K);	// output
	MECH_DDR |= (1<<MECH_PIN1)|(1<<MECH_PIN1);		// output

	// *** Настройка таймеров
	// TIMER0 - обеспечивает работу нагрузки, выдает импульсы на симистор (500uSec)
	// TIMER2 - системный счетчик, считает интервалы времени (1mSec)

	TCCR0 = (0 << CS02) | (1 << CS01) | (0 << CS00); // предделитель 8 при частоте 2МГц
	TCCR2 = (0 << CS22) | (1 << CS21) | (0 << CS20); // предделитель 8 при частоте 2МГц
	TCNT0 = TCNT0_BASE;		// счетчик таймера 0, 1 тик - 500 микросекунд
	TCNT2 = TCNT2_BASE;		// счетчик таймера 2, 1 тик - 1 миллисекунда
	TIMSK = (1 << TOIE0) | (1 << TOIE2);// разрешить прерывание таймеров по переполнению

	// *** Настройка исходных значений

	SetBit(SENSOR_DDR, SENSOR_A);// Устанавливаем PIN1 и PIN2 на выход
	SetBit(SENSOR_DDR, SENSOR_K);
	ClearBit(SENSOR_PORT, SENSOR_A);	// PIN1 -
	ClearBit(SENSOR_PORT, SENSOR_K);	// PIN2 -

	// *** Разрешаем прерывания
	sei();
}

// Получаем состояние сенсора
// Возвращает 1, если сенсор сработал (освещенность выше порога срабатывания), иначе 0
uint8_t GetSensorState(void)
{
	uint16_t counter = 0;

	for (uint8_t i = 0; i < 8; i++)	// Берем 4 отсчета.
	{
		// обеспечиваем обратное включение светодиода, заряжая паразитную ёмкость
		// ног микроконтроллера и светодиода
		SetBit(SENSOR_DDR, SENSOR_A); 	//настраиваем как выходы PIN1 и PIN2
		SetBit(SENSOR_DDR, SENSOR_K);	// OUTPUT

		SetBit(SENSOR_PORT, SENSOR_A);	// PIN1 +
		SetBit(SENSOR_PORT, SENSOR_K);	// PIN2 +	Разряжаем, что б снять заряд от предыдущих замеров
		asm("nop"); asm("nop"); asm("nop"); // небольшая задержка
		ClearBit(SENSOR_PORT, SENSOR_K);// PIN2 -	Заряжаем

		// изолируем PIN2 от светодиода
		ClearBit(SENSOR_DDR, SENSOR_K);	// PIN2 как INPUT
		ClearBit(SENSOR_PORT, SENSOR_K);// отключаем втроенный в МК подтягивающий резистор

		// считаем сколько требуется времени на разряд паразитной емкости до уровня логического нуля
		uint16_t j = 0;
		while (bit_is_clear(SENSOR_PIN, SENSOR_K))	j++;

		counter += j;	// Накапливаем отсчеты
	}

	counter = counter >> 4;	// И усредняем для устранения помех.

	// Проверяем, превысили ли порог освещенности
	return (counter < LIGHT_THRESHOLD) ? 1 : 0;
}

// Мигаем светодиодом-сенсором с учетом параметра BLINK_SPEED
void SensorBlink(void)
{
	// Мигаем светодиодом, показывая, что мы в работе
	SetBit(SENSOR_DDR, SENSOR_A); 		//настраиваем как выходы PIN1 и PIN2
	SetBit(SENSOR_DDR, SENSOR_K);		// OUTPUT
	ClearBit(SENSOR_PORT, SENSOR_A);	// PIN1 -
	ClearBit(SENSOR_PORT, SENSOR_K);	// PIN2 -

	blink_count++;

	if (blink_count > BLINK_SPEED)
		blink_count = 0;
	if (blink_count == BLINK_SPEED)
		SetBit(SENSOR_PORT, SENSOR_K);	// PIN1 +
}

// Таймер для обеспечения работы нагрузки,
// просто бомбит импульсами с периодом 500 мкс на управляющий вход симистора
ISR(TIMER0_OVF_vect)
{
	TCNT0 = TCNT0_BASE;

	if (LOAD_STATUS)
	{
		if (STATE_CTRL)
		{
			SetBit(MECH_PORT, MECH_PIN1);
			SetBit(MECH_PORT, MECH_PIN2);
		}
		else
		{
			ClearBit(MECH_PORT, MECH_PIN1);
			ClearBit(MECH_PORT, MECH_PIN2);
		}
		STATE_CTRL = STATE_CTRL ^ 0xFF;
	}
	else
	{
		ClearBit(MECH_PORT, MECH_PIN1);
		ClearBit(MECH_PORT, MECH_PIN2);
	}
}

// Системный таймер и таймер управления временем работы нагрузки
ISR(TIMER2_OVF_vect)
{
	++sys_time;
	TCNT2 = TCNT2_BASE;

	if (load_timer>0) {		// Если таймер нагрузки еще не отработал
		load_timer--;
		LOAD_STATUS = LOAD_ON;
	}
	else{					// Если таймер нагрузки уже отработал
		LOAD_STATUS = LOAD_OFF;
	}
}

